"use strict";
/**
 * @file upload files
 * @author nighca <nighca@live.cn>
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const walk_1 = __importDefault(require("walk"));
const qiniu_1 = __importDefault(require("qiniu"));
const logger_1 = __importDefault(require("./utils/logger"));
const build_conf_1 = require("./utils/build-conf");
const paths_1 = require("./utils/paths");
const utils_1 = require("./utils");
const logLifecycle = require('./utils').logLifecycle;
// 用于识别结果文件中的 sourcemap 文件
const sourceMapPattern = /\.map$/;
// https://gist.github.com/nighca/6562d098ac01814b6e1c1718b16d4abc
function batch(process, limit = -1) {
    return function batchProcess(tasks = []) {
        let results = [], finished = 0, current = 0;
        let rejected = false;
        function tryProcess(resolve, reject) {
            if (rejected)
                return;
            if (finished >= tasks.length)
                return resolve(results);
            if (current >= tasks.length)
                return;
            const index = current++;
            process(tasks[index]).then(result => {
                results[index] = result;
                finished++;
                tryProcess(resolve, reject);
            }, err => {
                reject(err);
                rejected = true;
            });
        }
        return new Promise((resolve, reject) => {
            const realLimit = limit > 0 ? limit : tasks.length;
            for (let i = 0; i < realLimit; i++) {
                tryProcess(resolve, reject);
            }
        });
    };
}
function getAllFiles(baseDir) {
    return new Promise((resolve, reject) => {
        const walker = walk_1.default.walk(baseDir);
        const files = [];
        walker.on('errors', (_, stat) => {
            reject(stat[0].error);
        });
        walker.on('nodeError', (_, stat) => {
            reject(stat[0].error);
        });
        walker.on('directoryError', (_, stat) => {
            reject(stat[0].error);
        });
        walker.on('files', (root, stats, next) => {
            stats.forEach(stat => files.push(path_1.default.relative(baseDir, path_1.default.join(root, stat.name))));
            next();
        });
        walker.on('end', () => {
            resolve(files);
        });
    });
}
function runWithRetry(process, retryTime = 3) {
    return __awaiter(this, void 0, void 0, function* () {
        let error = null;
        while (retryTime-- > 0) {
            try {
                return yield process();
            }
            catch (e) {
                error = e;
            }
        }
        throw error;
    });
}
function uploadFile(localFile, bucket, key, mac) {
    return __awaiter(this, void 0, void 0, function* () {
        const options = {
            scope: bucket + ':' + key
        };
        const putPolicy = new qiniu_1.default.rs.PutPolicy(options);
        const uploadToken = putPolicy.uploadToken(mac);
        const putExtra = new qiniu_1.default.form_up.PutExtra();
        const config = new qiniu_1.default.conf.Config();
        const formUploader = new qiniu_1.default.form_up.FormUploader(config);
        const putFile = () => new Promise((resolve, reject) => {
            formUploader.putFile(uploadToken, key, localFile, putExtra, (err, ret) => {
                if (err || ret.error) {
                    reject(err || ret.error);
                    return;
                }
                resolve();
            });
        });
        return runWithRetry(putFile, 3);
    });
}
function upload() {
    return __awaiter(this, void 0, void 0, function* () {
        const buildConfig = yield build_conf_1.findBuildConfig();
        const deployConfig = buildConfig.deploy.config;
        const distPath = paths_1.getDistPath(buildConfig);
        const prefix = utils_1.getPathFromUrl(buildConfig.publicUrl, false);
        const mac = new qiniu_1.default.auth.digest.Mac(deployConfig.accessKey, deployConfig.secretKey);
        const files = yield getAllFiles(distPath);
        const concurrentLimit = 50;
        const batchUploadFile = batch((name) => __awaiter(this, void 0, void 0, function* () {
            const key = prefix ? `${prefix}/${name}` : name;
            const filePath = path_1.default.resolve(distPath, name);
            // 排除 sourcemap 文件，不要上传到生产环境 CDN
            if (sourceMapPattern.test(filePath)) {
                return logger_1.default.info(`[IGNORE] ${filePath}`);
            }
            yield uploadFile(filePath, deployConfig.bucket, key, mac);
            logger_1.default.info(`[UPLOAD] ${filePath} -> ${key}`);
        }), concurrentLimit);
        yield batchUploadFile(files);
        logger_1.default.info(`[UPLOAD] ${files.length} files with concurrent limit ${concurrentLimit}`);
    });
}
exports.default = logLifecycle('Upload', upload, logger_1.default);
