"use strict";
/**
 * @file util methods
 * @author nighca <nighca@live.cn>
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.watchFile = exports.getPageFilename = exports.makeSnapshotGlob = exports.makeTestGlob = exports.makeTestRegexpText = exports.removeSuffix = exports.isImage = exports.getPathFromUrl = exports.logLifecycle = exports.extend = void 0;
const fs_1 = __importDefault(require("fs"));
const lodash_1 = require("lodash");
const url_1 = require("url");
function extend(...addons) {
    const target = {};
    addons.forEach(addon => {
        for (let key in addon) {
            if (addon.hasOwnProperty(key)) {
                target[key] = addon[key];
            }
        }
    });
    return target;
}
exports.extend = extend;
const durationHumanizeSteps = [
    { unit: 'day', amount: 1000 * 60 * 60 * 24 },
    { unit: 'hour', amount: 1000 * 60 * 60 },
    { unit: 'min', amount: 1000 * 60 },
    { unit: 's', amount: 1000 },
    { unit: 'ms', amount: 1 }
];
function getDurationFrom(startAt) {
    const duration = Date.now() - startAt;
    if (duration < 1) {
        return '0ms';
    }
    const result = durationHumanizeSteps.reduce(([parts, duration], { unit, amount }) => {
        if (duration >= amount) {
            parts.push(`${Math.floor(duration / amount)}${unit}`);
            duration = duration % amount;
        }
        return [parts, duration];
    }, [[], duration]);
    return result[0].join(' ');
}
/** log behavior lifecycle */
function logLifecycle(
/** behavior name */
name, 
/** target method */
method, logger) {
    return (...args) => __awaiter(this, void 0, void 0, function* () {
        logger.info(`${name} start`);
        const startAt = Date.now();
        try {
            const result = yield method(...args);
            logger.info(`${name} succeeded, costs: ${getDurationFrom(startAt)}`);
            return result;
        }
        catch (e) {
            logger.error(`${name} failed:`, e);
            throw e;
        }
    });
}
exports.logLifecycle = logLifecycle;
/** get prefix part from url */
function getPathFromUrl(targetUrl, 
/** if with slash at begining & ending */
withSlash = true) {
    // 支持一下 `./` 形式的 url，暂不考虑 `../` 的情况
    targetUrl = targetUrl.replace(/^\.\//, '\/');
    const parsed = url_1.parse(targetUrl);
    const pathWithoutSlash = (parsed.pathname || '').replace(/^\/|\/$/g, '');
    return (withSlash
        ? (pathWithoutSlash ? `/${pathWithoutSlash}/` : '/')
        : pathWithoutSlash);
}
exports.getPathFromUrl = getPathFromUrl;
/** check isImage by extension */
function isImage(extension) {
    // svg 有可能是字体文件，且 svg 作为图片时本身压缩空间小，暂不做处理
    return ['png', 'jpg', 'jpeg', 'gif'].some(type => type === extension);
}
exports.isImage = isImage;
/** remove suffix in filePath */
function removeSuffix(filePath) {
    return filePath.replace(/\.\w+$/, '');
}
exports.removeSuffix = removeSuffix;
function escapeRegExp(str) {
    return str.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&");
}
/** make regexp test for test files */
function makeTestRegexpText(extensions) {
    return `\\.spec\\.(${extensions.map(escapeRegExp).join('|')})$`;
}
exports.makeTestRegexpText = makeTestRegexpText;
/** make glob for test files (under build root) */
function makeTestGlob(srcDir, extensions) {
    if (extensions.length === 0) {
        throw new Error('at least one extension required');
    }
    if (extensions.length === 1) {
        const extension = extensions[0];
        return `${srcDir}/**/*.spec.${extension}`;
    }
    const extensionsText = extensions.join('|');
    return `${srcDir}/**/*.spec.@(${extensionsText})`;
}
exports.makeTestGlob = makeTestGlob;
/** make glob for snapshot glob (under build root) */
function makeSnapshotGlob(srcDir) {
    return `${srcDir}/**/*.snap`;
}
exports.makeSnapshotGlob = makeSnapshotGlob;
/** 从 page 名（buildConfig.pages 的 key）得到在 serve / 构建时对应的页面文件名  */
function getPageFilename(pageName) {
    return `${pageName}.html`;
}
exports.getPageFilename = getPageFilename;
/** 监听文件内容变化，会对内容做比对，适用于体积较小的文本文件 */
function watchFile(filePath, listener) {
    function readFile() {
        return fs_1.default.readFileSync(filePath, { encoding: 'utf-8' });
    }
    let previousCnt = readFile();
    // 1. debounce 是因为有时候可能会在短时间内触发多次变更事件，这里做下合并
    // 2. 比较文件内容是因为触发变更事件时文件的内容可能没有其实没有变化，这里假设使用方只关心内容的变化
    const onChange = lodash_1.debounce(() => {
        const currentCnt = readFile();
        if (previousCnt !== currentCnt) {
            previousCnt = currentCnt;
            listener(currentCnt);
        }
    }, 100);
    const fsWatcher = fs_1.default.watch(filePath);
    fsWatcher.on('change', onChange);
    return () => fsWatcher.close();
}
exports.watchFile = watchFile;
