"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ignoreWarning = exports.enableFilesystemCache = exports.addDefaultExtension = exports.makeRule = exports.appendCacheGroups = exports.parseOptimizationConfig = exports.processSourceMap = exports.adaptLoader = exports.appendPlugins = void 0;
const immer_1 = require("immer");
const path_1 = __importDefault(require("path"));
const chunks_1 = __importDefault(require("../constants/chunks"));
const paths_1 = require("./paths");
const logger_1 = __importDefault(require("./logger"));
/** 向配置中追加 plugin */
function appendPlugins(config, ...plugins) {
    return immer_1.produce(config, newConfig => {
        newConfig.plugins = newConfig.plugins || [];
        for (const plugin of plugins) {
            if (plugin != null) {
                newConfig.plugins.push(plugin);
            }
        }
    });
}
exports.appendPlugins = appendPlugins;
/** 对 loader 信息进行标准化处理 */
function adaptLoader({ loader, options }) {
    // 这里先前会把 loader 的值通过 `require.resolve` 替换为绝对路径
    // 目的是为了保证总是能拿到正确的一致的 loader 实现（不受项目本地 node_modules 中 loader 实现的影响）
    const resolvedLoader = require.resolve(loader);
    const loaderObj = { loader: resolvedLoader, options };
    // 添加 builder 后续处理需要的字段，`enumerable: false` 是为了防止 webpack 乱报错
    // 比如 extract-style 处要根据 `__loaderName__ === 'style-loader'` 判断是否需处理
    Object.defineProperty(loaderObj, '__loaderName__', {
        value: loader,
        enumerable: false,
        configurable: true
    });
    return loaderObj;
}
exports.adaptLoader = adaptLoader;
const sourcemapParseFailedWarningPattern = /Failed to parse source map from/;
/** 配置 source map */
function processSourceMap(config, highQuality) {
    config = immer_1.produce(config, (newConfig) => {
        // 使用 cheap-module-source-map 而不是 eval-cheap-module-source-map 或 eval-source-map
        // 具体原因见 https://github.com/Front-End-Engineering-Cloud/builder/pull/139#discussion_r676475522
        newConfig.devtool = highQuality ? 'cheap-module-source-map' : 'eval';
        newConfig.module.rules.push({
            test: /node_modules\/.*\.js$/,
            enforce: 'pre',
            use: [{
                    loader: 'source-map-loader',
                    options: {
                        filterSourceMappingUrl(_url, _resourcePath) {
                            // highQuality 为 false 时也需要引入 source-map-loader 并在这里返回 remove
                            // 来把第三方库代码中的 SourceMappingURL 信息干掉，以避免开发时的 warning
                            return highQuality ? 'consume' : 'remove';
                        }
                    }
                }].map(adaptLoader)
        });
    });
    if (highQuality) {
        // 第三方依赖的 source map 信息残缺导致 parse 失败是很常见的事情，
        // 最常见的原因是生成的 sourcemap 未将源码 inline，而发布的内容里也没有包含源码，
        // 详见 https://www.typescriptlang.org/tsconfig/#inlineSources
        // 这里忽略相关 warning，避免污染浏览器 console
        config = ignoreWarning(config, sourcemapParseFailedWarningPattern);
    }
    return config;
}
exports.processSourceMap = processSourceMap;
/** 解析 optimization 配置，获取 chunks 及 cacheGroups */
function parseOptimizationConfig(optimization) {
    const { extractVendor, extractCommon } = optimization;
    const baseChunks = [];
    const cacheGroups = {};
    if (extractVendor) {
        if (typeof extractVendor === 'string') {
            throw new Error('BREAKING CHANGE: extractVendor 已不再支持该用法，使用方式请参考帮助文档！');
        }
        else if (typeof extractVendor === 'boolean' || extractVendor.length > 0) {
            baseChunks.push(chunks_1.default.vendor);
            cacheGroups[chunks_1.default.vendor] = {
                name: chunks_1.default.vendor,
                chunks: 'initial',
                priority: -10,
                test: function (module) {
                    const resource = module.resource;
                    if (!resource)
                        return false;
                    const nodeModulesPath = path_1.default.join(path_1.default.sep, 'node_modules', path_1.default.sep);
                    if (typeof extractVendor === 'boolean') {
                        return resource.includes(nodeModulesPath);
                    }
                    return extractVendor.some(packageName => {
                        return resource.includes(path_1.default.join(nodeModulesPath, packageName, path_1.default.sep));
                    });
                }
            };
        }
    }
    if (extractCommon) {
        baseChunks.push(chunks_1.default.common);
        cacheGroups[chunks_1.default.common] = {
            name: chunks_1.default.common,
            chunks: 'initial',
            minChunks: 2,
            priority: -20
        };
    }
    return { baseChunks, cacheGroups };
}
exports.parseOptimizationConfig = parseOptimizationConfig;
/** 向配置中追加 cacheGroup 项 */
function appendCacheGroups(config, cacheGroups) {
    return immer_1.produce(config, newConfig => {
        var _a, _b;
        newConfig.optimization || (newConfig.optimization = {});
        (_a = newConfig.optimization).splitChunks || (_a.splitChunks = {});
        (_b = newConfig.optimization.splitChunks).cacheGroups || (_b.cacheGroups = {});
        Object.assign(newConfig.optimization.splitChunks.cacheGroups, cacheGroups);
    });
}
exports.appendCacheGroups = appendCacheGroups;
/** 构造 rule */
function makeRule(extension, resource, context, base) {
    const issuerConditions = [];
    if (context.include != null)
        issuerConditions.push(context.include);
    if (context.exclude.length > 0)
        issuerConditions.push({ not: context.exclude });
    const rule = Object.assign(Object.assign({}, base), { test: resource.include, exclude: resource.exclude, issuer: issuerConditions.length > 0 ? { and: issuerConditions } : undefined });
    // 添加 builder 后续处理需要的字段，`enumerable: false` 是为了防止 webpack 乱报错
    Object.defineProperty(rule, '__extension__', {
        value: extension,
        enumerable: false,
        configurable: true
    });
    return rule;
}
exports.makeRule = makeRule;
/** 添加默认文件后缀名 */
function addDefaultExtension(config, extension) {
    extension = extension && ('.' + extension);
    return immer_1.produce(config, newConfig => {
        const resolve = newConfig.resolve = newConfig.resolve || {};
        const extensions = resolve.extensions = resolve.extensions || [];
        if (!extensions.includes(extension)) {
            extensions.push(extension);
        }
    });
}
exports.addDefaultExtension = addDefaultExtension;
/** 开启文件缓存 */
function enableFilesystemCache(config) {
    return immer_1.produce(config, newConfig => {
        newConfig.cache = {
            type: 'filesystem',
            cacheDirectory: paths_1.getCachePath()
        };
    });
}
exports.enableFilesystemCache = enableFilesystemCache;
/** 指定 pattern 忽略 warning */
function ignoreWarning(config, pattern) {
    return immer_1.produce(config, newConfig => {
        var _a;
        (_a = newConfig.ignoreWarnings) !== null && _a !== void 0 ? _a : (newConfig.ignoreWarnings = []);
        if (!newConfig.ignoreWarnings.includes(pattern)) {
            logger_1.default.debug('append warningsFilter:', pattern);
            newConfig.ignoreWarnings.push(pattern);
        }
    });
}
exports.ignoreWarning = ignoreWarning;
