"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getServeConfig = exports.getConfig = void 0;
const lodash_1 = require("lodash");
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const webpack_1 = require("webpack");
const html_webpack_plugin_1 = __importDefault(require("html-webpack-plugin"));
const copy_webpack_plugin_1 = __importDefault(require("copy-webpack-plugin"));
const react_refresh_webpack_plugin_1 = __importDefault(require("@pmmmwh/react-refresh-webpack-plugin"));
const mini_css_extract_plugin_1 = __importDefault(require("mini-css-extract-plugin"));
const css_minimizer_webpack_plugin_1 = __importDefault(require("css-minimizer-webpack-plugin"));
const webpack_bundle_analyzer_1 = require("webpack-bundle-analyzer");
const webpackbar_1 = __importDefault(require("webpackbar"));
const image_minimizer_webpack_plugin_1 = __importDefault(require("image-minimizer-webpack-plugin"));
const paths_1 = require("../utils/paths");
const build_conf_1 = require("../utils/build-conf");
const transform_1 = require("./transform");
const build_env_1 = require("../utils/build-env");
const logger_1 = __importDefault(require("../utils/logger"));
const utils_1 = require("../utils");
const webpack_2 = require("../utils/webpack");
const svgo_1 = require("../utils/svgo");
const dirnameOfBuilder = path_1.default.resolve(__dirname, '../..');
const nodeModulesOfBuilder = path_1.default.resolve(dirnameOfBuilder, 'node_modules');
/** 获取 webpack 配置（构建用） */
function getConfig() {
    return __awaiter(this, void 0, void 0, function* () {
        const buildConfig = yield build_conf_1.findBuildConfig();
        const isProd = build_env_1.getEnv() === build_env_1.Env.Prod;
        const isDev = build_env_1.getEnv() === build_env_1.Env.Dev;
        const resolveAlias = lodash_1.mapValues(buildConfig.resolve.alias, path => paths_1.abs(path));
        let config = {
            target: 'web',
            mode: getMode(),
            context: paths_1.getBuildRoot(),
            resolve: {
                // 同默认配置，这里写出来是因为后续会有新增 extensions
                extensions: ['.wasm', '.mjs', '.js', '.json'],
                modules: [
                    paths_1.getSrcPath(buildConfig),
                    'node_modules',
                    nodeModulesOfBuilder,
                    paths_1.abs('node_modules')
                ],
                alias: resolveAlias
            },
            resolveLoader: {
                modules: [
                    'node_modules',
                    nodeModulesOfBuilder
                ]
            },
            entry: lodash_1.mapValues(buildConfig.entries, entryFile => paths_1.abs(entryFile)),
            module: { rules: [] },
            plugins: [],
            output: {
                path: paths_1.getDistPath(buildConfig),
                filename: 'static/[name]-[contenthash].js',
                chunkFilename: 'static/[id]-[chunkhash].js',
                assetModuleFilename: 'static/[name]-[contenthash][ext]',
                publicPath: (isProd
                    ? buildConfig.publicUrl
                    : utils_1.getPathFromUrl(buildConfig.publicUrl)),
                environment: {
                    // 这里控制 webpack 本身的运行时代码（而不是业务代码），
                    // 对于语言 feature 先全部配合不支持，以确保 webpack 会产出兼容性最好的代码；
                    // TODO: 后续考虑通过使用 build config 中的 targets.browsers 来挨个判断是否支持
                    arrowFunction: false,
                    bigIntLiteral: false,
                    const: false,
                    destructuring: false,
                    dynamicImport: false,
                    forOf: false,
                    module: false
                }
            },
            optimization: {
                minimizer: [
                    '...',
                    new css_minimizer_webpack_plugin_1.default()
                ]
            }
        };
        let baseChunks = [];
        if (isProd) {
            const result = webpack_2.parseOptimizationConfig(buildConfig.optimization);
            baseChunks = result.baseChunks;
            config = webpack_2.appendCacheGroups(config, result.cacheGroups);
        }
        if (isDev) {
            config = webpack_2.processSourceMap(config, buildConfig.optimization.highQualitySourceMap);
        }
        config = transform_1.addTransforms(config, buildConfig);
        const htmlPlugins = Object.entries(buildConfig.pages).map(([name, { template, entries }]) => {
            return new html_webpack_plugin_1.default({
                template: paths_1.abs(template),
                filename: utils_1.getPageFilename(name),
                chunks: [...baseChunks, ...entries],
                chunksSortMode: 'manual'
            });
        });
        const definePlugin = new webpack_1.DefinePlugin(
        // webpack DefinePlugin 只是简单的文本替换，这里进行 JSON stringify 转换
        lodash_1.mapValues(Object.assign({ 'process.env.NODE_ENV': build_env_1.getEnv() }, buildConfig.envVariables), JSON.stringify));
        const staticDirCopyPlugin = getStaticDirCopyPlugin(buildConfig);
        config = webpack_2.appendPlugins(config, ...htmlPlugins, definePlugin, staticDirCopyPlugin, new webpackbar_1.default({ color: 'green' }));
        if (isProd) {
            config = webpack_2.appendPlugins(config, new mini_css_extract_plugin_1.default({
                filename: 'static/[name]-[contenthash].css',
                chunkFilename: 'static/[id]-[chunkhash].css'
            }));
        }
        if (build_conf_1.getNeedAnalyze()) {
            config = webpack_2.appendPlugins(config, new webpack_bundle_analyzer_1.BundleAnalyzerPlugin());
        }
        if (isProd && buildConfig.optimization.compressImage) {
            config = webpack_2.appendPlugins(config, new image_minimizer_webpack_plugin_1.default({
                minimizerOptions: {
                    plugins: [
                        ['mozjpeg', { progressive: true, quality: 65 }],
                        ['gifsicle', { interlaced: false }],
                        ['svgo', svgo_1.svgoConfigForImagemin]
                        // 这里先不做 png 的压缩，因为 imagemin-pngquant 有可能会产生负优化（结果文件比源文件体积大），
                        // 而且目前不支持 option 来在产生负优化时直接使用源文件，相关 issue https://github.com/kornelski/pngquant/issues/338
                        // ['pngquant', { quality: [0.65, 0.9], speed: 4 }],
                    ]
                }
            }));
        }
        if (paths_1.getNeedCache()) {
            config = webpack_2.enableFilesystemCache(config);
        }
        return config;
    });
}
exports.getConfig = getConfig;
/** 获取 webpack 配置（dev server 用，不含 dev server 配置） */
function getServeConfig() {
    return __awaiter(this, void 0, void 0, function* () {
        const config = yield getConfig();
        return webpack_2.appendPlugins(config, new react_refresh_webpack_plugin_1.default());
    });
}
exports.getServeConfig = getServeConfig;
/** 获取合适的 webpack mode */
function getMode() {
    const buildEnv = build_env_1.getEnv();
    if (buildEnv === build_env_1.Env.Dev)
        return 'development';
    if (buildEnv === build_env_1.Env.Prod)
        return 'production';
    return 'none';
}
/** 构造用于 static 目录复制的 plugin 实例 */
function getStaticDirCopyPlugin(buildConfig) {
    const staticPath = paths_1.getStaticPath(buildConfig);
    if (!fs_1.default.existsSync(staticPath))
        return null;
    try {
        const stats = fs_1.default.statSync(staticPath);
        if (!stats.isDirectory()) {
            throw new Error('staticPath not a directory');
        }
        return new copy_webpack_plugin_1.default({
            patterns: [{ from: staticPath, to: 'static', toType: 'dir' }]
        });
    }
    catch (e) {
        logger_1.default.warn('Copy staticDir content failed:', e && e.message);
    }
}
