"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.addTransforms = void 0;
const immer_1 = __importDefault(require("immer"));
const postcss_preset_env_1 = __importDefault(require("postcss-preset-env"));
const mini_css_extract_plugin_1 = __importDefault(require("mini-css-extract-plugin"));
const transform_1 = require("../constants/transform");
const svgo_1 = require("../utils/svgo");
const build_conf_1 = require("../utils/build-conf");
const build_env_1 = require("../utils/build-env");
const webpack_1 = require("../utils/webpack");
// ts-loader 开启 transpileOnly 后会出的 warning
const tsTranspileOnlyWarningPattern = /export .* was not found in/;
/** 添加资源转换逻辑（buildConfig.transforms） */
function addTransforms(
/** 当前 webpack 配置 */
config, 
/** 构建配置 build config */
buildConfig) {
    const transformConfigs = Object.entries(buildConfig.transforms).map(([condition, transform]) => {
        const [extensionValue, contextValue = ''] = condition.split('@');
        const resource = { include: extensionValue, exclude: [] };
        const context = { include: contextValue, exclude: [] };
        return { transform, resource, context };
    });
    // 处理 extension / context 冲突的 transform 项
    transformConfigs.forEach(transformConfig => {
        const { resource, context } = transformConfig;
        // 若存在其他 extension 比当前 extension 更精确，则在当前项中排除之；
        // 如 css & module.css，则在 css 对应项中排除 module.css
        const extensionValues = transformConfigs.map(({ resource }) => resource.include);
        resource.exclude = extensionValues.filter(target => target !== resource.include && endsWithExt(target, resource.include));
        // 若存在其他项 extension 与当前 extension 相同，而 context 更精确，则在当前项中排除之；
        // 如 svg@css & svg@module.css，则在 svg@css 对应项中排除 svg@module.css
        const contextValues = transformConfigs.filter(targetConfig => targetConfig.resource.include === resource.include).map(({ context }) => context.include);
        context.exclude = contextValues.filter(target => target !== context.include && endsWithExt(target, context.include));
    });
    transformConfigs.forEach(({ transform, resource, context }) => {
        config = addTransform(config, buildConfig, transform, resource, context);
    });
    return config;
}
exports.addTransforms = addTransforms;
function addTransform(
/** 当前 webpack 配置 */
config, 
/** 构建配置 build config */
{ targets, optimization }, 
/** transform 信息 */
transform, 
/** 资源后缀名条件 */
resource, 
/** 上下文资源（引入当前资源的资源）后缀名条件 */
context) {
    // resource.include 当前处理的文件后缀
    const extension = resource.include;
    const excludePatterns = resource.exclude.map(makeExtensionPattern);
    // 针对后缀为 js 的 transform，控制范围（不对依赖做转换）
    if (extension === 'js') {
        const jsExcludePattern = makeJsExcludePattern(optimization.transformDeps);
        if (jsExcludePattern != null)
            excludePatterns.push(jsExcludePattern);
    }
    const resourcePattern = {
        include: makeExtensionPattern(extension),
        exclude: excludePatterns
    };
    const contextPattern = {
        include: context.include ? makeExtensionPattern(context.include) : undefined,
        exclude: context.exclude.map(makeExtensionPattern)
    };
    const appendRule = (previousConfig, ruleBase) => immer_1.default(previousConfig, (nextConfig) => {
        const rule = webpack_1.makeRule(extension, resourcePattern, contextPattern, ruleBase);
        nextConfig.module.rules.push(rule);
    });
    const appendRuleWithLoaders = (previousConfig, ...loaders) => (appendRule(previousConfig, { use: loaders.map(webpack_1.adaptLoader) }));
    const appendRuleWithAssetType = (previousConfig, assetType) => (appendRule(previousConfig, { type: assetType }));
    const markDefaultExtension = (previousConfig) => {
        return webpack_1.addDefaultExtension(previousConfig, extension);
    };
    switch (transform.transformer) {
        case transform_1.Transform.Css:
        case transform_1.Transform.Less: {
            const transformConfig = (transform.config || {});
            const loaders = [];
            if (build_env_1.getEnv() === build_env_1.Env.Dev) {
                loaders.push({ loader: 'style-loader' });
            }
            else if (build_env_1.getEnv() === build_env_1.Env.Prod) {
                // dev 环境不能用 MiniCssExtractPlugin.loader
                // 已知 less with css-module 的项目，样式 hot reload 会有问题
                loaders.push({ loader: mini_css_extract_plugin_1.default.loader });
            }
            loaders.push({
                loader: 'css-loader',
                options: {
                    // https://github.com/webpack-contrib/css-loader/issues/228#issuecomment-312885975
                    importLoaders: transform.transformer === transform_1.Transform.Css ? 1 : 0,
                    modules: (transformConfig.modules
                        ? { localIdentName: '[local]_[hash:base64:5]' }
                        : false)
                }
            });
            loaders.push({ loader: 'postcss-loader', options: {
                    postcssOptions: {
                        plugins: [
                            postcss_preset_env_1.default({
                                browsers: targets.browsers.join(', ')
                            })
                        ]
                    }
                } });
            if (transform.transformer === transform_1.Transform.Less) {
                loaders.push({
                    loader: 'less-loader',
                    options: {
                        lessOptions: transformConfig.options
                    }
                });
            }
            return appendRuleWithLoaders(config, ...loaders);
        }
        case transform_1.Transform.Babel: {
            config = markDefaultExtension(config);
            const transformConfig = (transform.config || {});
            return appendRuleWithLoaders(config, {
                loader: 'babel-loader',
                options: makeBabelLoaderOptions(transformConfig, targets.browsers, optimization.addPolyfill)
            });
        }
        case transform_1.Transform.Jsx: {
            config = markDefaultExtension(config);
            const transformConfig = (transform.config || {});
            return appendRuleWithLoaders(config, {
                loader: 'babel-loader',
                options: makeBabelLoaderOptions(transformConfig.babelOptions || {}, targets.browsers, optimization.addPolyfill, true)
            });
        }
        case transform_1.Transform.Ts:
        case transform_1.Transform.Tsx: {
            config = markDefaultExtension(config);
            const transformConfig = Object.assign({ transpileOnlyWhenDev: true, babelOptions: {} }, transform.config);
            const babelOptions = makeBabelLoaderOptions(transformConfig.babelOptions, targets.browsers, optimization.addPolyfill, transform.transformer === transform_1.Transform.Tsx);
            const compilerOptions = {
                // 这里设置为 ES2020（最新的规范能力），需要注意的是，这里设置 ESNext 可能是不合适的：
                // 
                // > The special ESNext value refers to the highest version your version of TypeScript supports. This setting should be used with caution, 
                // > since it doesn’t mean the same thing between different TypeScript versions and can make upgrades less predictable.
                // > - https://www.typescriptlang.org/tsconfig#target
                // 
                // 这里 Typescript 处理的结果会交给 babel 处理；我们默认使用 @babel/preset-env，预期会支持最新的规范能力
                // 然而我们使用的 Typescript 跟 babel (& @babel/preset-env) 行为之间可能会有 gap：
                // 以 babel-plugin-proposal-class-properties 为例，在对应的 proposal 进入 stage 4 后，
                // Typescript 会认为以 ESNext 为目标时，对应的语法不再需要转换；
                // 而如果 builder 此时依赖了相对更新的 Typescript 版本，以及相对较旧的 babel (& @babel/preset-env) 版本
                // 那么这里对 class properties 语法的支持就会有问题（Typescript & babel 都不会对它进行转换）
                target: 'ES2020',
                // 跟 target 保持一致，而不是设置为 CommonJS；由 webpack 来做 module 格式的转换以 enable tree shaking
                module: 'ES2020',
                // module 为 ES2020 时，moduleResolution 默认为 Classic（虽然 TS 文档不是这么说的），这里明确指定为 Node
                moduleResolution: 'Node'
            };
            const tsLoaderOptions = {
                transpileOnly: build_env_1.getEnv() === build_env_1.Env.Dev && transformConfig.transpileOnlyWhenDev,
                compilerOptions,
                // 方便项目直接把内部依赖（portal-base / fe-core 等）的源码 link 进来一起构建调试
                allowTsInNodeModules: true
            };
            if (tsLoaderOptions.transpileOnly) {
                // 干掉因为开启 transpileOnly 导致的 warning
                // 详情见 https://github.com/TypeStrong/ts-loader#transpileonly
                config = webpack_1.ignoreWarning(config, tsTranspileOnlyWarningPattern);
            }
            return appendRuleWithLoaders(config, { loader: 'babel-loader', options: babelOptions }, 
            // 这边预期 ts-loader 将 ts 代码编成 ES6 代码，然后再交给 babel-loader 处理
            { loader: 'ts-loader', options: tsLoaderOptions });
        }
        case transform_1.Transform.Raw: {
            return appendRuleWithAssetType(config, 'asset/source');
        }
        case transform_1.Transform.File: {
            return appendRuleWithAssetType(config, 'asset');
        }
        case transform_1.Transform.SvgSprite: {
            throw new Error('Transform svg-sprite is not supported any more.');
        }
        case transform_1.Transform.Svgr: {
            const shouldCompressSvg = build_env_1.getEnv() === build_env_1.Env.Prod && optimization.compressImage;
            const svgoConfig = svgo_1.getSvgoConfigForSvgr(shouldCompressSvg);
            return appendRuleWithLoaders(config, {
                loader: '@svgr/webpack',
                options: { svgo: true, svgoConfig }
            });
        }
        default: {
            throw new Error(`Invalid transformer: ${transform.transformer}`);
        }
    }
}
function endsWithExt(target, ext) {
    return target.endsWith(ext ? ('.' + ext) : '');
}
function regexpEscape(s) {
    return s.replace(/[-\/\\^$*+?.()|[\]{}]/g, '\\$&');
}
function makeExtensionPattern(ext) {
    return new RegExp(`\\.${regexpEscape(ext)}\$`);
}
/** 构造处理 Javascript 内容时排除依赖用的正则 */
function makeJsExcludePattern(
/** 是否排除依赖，或需要被处理（不应该被排除）的依赖包名 */
transformDeps) {
    if (Array.isArray(transformDeps)) {
        return new RegExp(`node_modules/(?!(${transformDeps.join('|')})/).*`);
    }
    return transformDeps ? null : /node_modules\//;
}
// 不支持 preset 简写的形式
function adaptBabelPreset(preset) {
    if (typeof preset === 'string') {
        return require.resolve(preset);
    }
    const [name, ...options] = preset;
    return [require.resolve(name), ...options];
}
// TODO: 添加 babel-plugin- 前缀
function adaptBabelPluginName(name) {
    return require.resolve(name);
}
function adaptBabelPlugin(plugin) {
    if (typeof plugin === 'string') {
        return adaptBabelPluginName(plugin);
    }
    const [name, ...options] = plugin;
    return [adaptBabelPluginName(name), ...options];
}
function includes(list, name) {
    return list.some(item => {
        const itemName = typeof item === 'string' ? item : item[0];
        return itemName === name;
    });
}
const corejsOptions = {
    version: 3,
    proposals: false
};
function getBabelPresetEnvOptions(targets, polyfill) {
    return Object.assign({ 
        // enable tree-shaking，由 webpack 来做 module 格式的转换
        modules: false, targets }, (
    // global polyfill
    build_conf_1.shouldAddGlobalPolyfill(polyfill)
        && {
            // https://babeljs.io/docs/en/babel-preset-env#usebuiltins
            useBuiltIns: 'usage',
            corejs: corejsOptions
        }));
}
/**
 * 构造 babel-loader 的配置对象，主要是添加默认的 polyfill 相关配置
 * 另外会调整 preset、plugin 的名字为绝对路径
 */
function makeBabelLoaderOptions(
/** babel options */
options, 
/** babel env targets: https://babeljs.io/docs/en/babel-preset-env#targets */
targets, 
/** polyfill 模式 */
polyfill, 
/** 是否 react 项目 */
withReact = false) {
    options = options || {};
    const isDev = build_env_1.getEnv() === build_env_1.Env.Dev;
    return immer_1.default(options, nextOptions => {
        const presets = nextOptions.presets || [];
        const presetEnvName = '@babel/preset-env';
        if (!isDev && !includes(presets, presetEnvName)) {
            presets.unshift([presetEnvName, getBabelPresetEnvOptions(targets, polyfill)]);
        }
        const presetReactName = '@babel/preset-react';
        if (withReact && !includes(presets, presetReactName)) {
            presets.push([presetReactName, { development: isDev }]);
        }
        nextOptions.presets = presets.map(adaptBabelPreset);
        const plugins = nextOptions.plugins || [];
        const pluginTransformRuntimeName = '@babel/plugin-transform-runtime';
        if (!isDev && build_conf_1.shouldAddRuntimePolyfill(polyfill) && !includes(plugins, pluginTransformRuntimeName)) {
            plugins.unshift([pluginTransformRuntimeName, { corejs: corejsOptions }]);
        }
        const pluginReactRefreshName = 'react-refresh/babel';
        if (withReact && isDev && !includes(plugins, pluginReactRefreshName)) {
            plugins.push(pluginReactRefreshName);
        }
        nextOptions.plugins = plugins.map(adaptBabelPlugin);
        // 用于指定预期模块类型，若用户未指定，则使用默认值 unambiguous，即：自动推断
        nextOptions.sourceType = nextOptions.sourceType || 'unambiguous';
    });
}
